// Program generates code for the calculation of the solution
// of a symmetrical set of linear equations Ax=b for a given dimension.

// The resulting code uses a Cholesky decomposition for fast computation.
// This code can also be used to obtain the inverse of the matrix, by solving 
// the set of linear equations several times for b each base vector.
//
// x and b are arrays of n elements. A is a nxn symmetrical matrix. 
// It is however represented as an array. Due to symmetry, the elements
// of the upper triangle are sufficient. The elements of this triangle are 
// put in the array from left to right and then from top to bottom.
//
// ex. for 3x3 matrix: [A1 A2 A3
//						A4 A5 A6
//						A7 A8 A9]
// 
//
// A = [A1 A2 A3 A5 A6 A9]
//  
//
// Change integer n according to desired dimension.
// Change string fileName according to desired... fileName.
//
// At the end of the generated code, the number of required arithmetical 
// operations is given.
//
//
// (c) Joris Naudet
// 19th May 2004

#include <iostream>
#include <stdlib.h>
#include <sstream>
#include "generateLinEquSolver.h"
#include <fstream>

int main() {


	//---------------------------------	
		int n=8;	// Dimension
		string fileName="test4x4.cpp";
	//---------------------------------


	string output;
	ofstream fileS(fileName.c_str());

	output=solveLinearEquations(n);

	fileS << output;

	fileS << endl << endl << "// *: " << numberMultiplications(output) << endl;
	fileS << "// /: " << numberDivisions(output) << endl;
	fileS << "// +/-: " << numberAdditions(output) << endl << endl;

	fileS.close();

	return(0);
}


string solveLinearEquations(int n) 
{
	ostringstream outStream;

	intMatrix indexD,indexL,indexM;
	int i,j,k,l;

	// Function header and declarations

	outStream << "//" << endl << "// Function solves a " << n << "x" << n
		<< " symmetrical set of linear equations," << endl << "// using Cholesky" 
		<< " decomposition" << endl << "//" << endl << endl;

	outStream << "void " << "linEqu" << n << "x" << n 
		<< "(double *x,double *m,double *b)" 
		<< endl << "{" << endl << "double invD[" << n << "],dl["
		<< (n+1)*n/2.0 << "],l[" << (n-1)*n/2.0 << "],g[" << n << "];" << endl
		<< endl;


	// Initialize vectors

	indexL.resize(n);
	indexM.resize(n);
	indexD.resize(n);
	for (i=0;i<n;i++) {
		indexL[i].resize(n);
		indexD[i].resize(n);
		indexM[i].resize(n);
	}

	// Fill indexMatrices

	k=0;l=0;
	for (i=0;i<n;i++) {
		for (j=i;j<n;j++) {
			indexM[i][j]=k;
			indexM[j][i]=k;
			indexD[i][j]=k;
			indexD[j][i]=k;
			k++;
		}
	}
	for (i=0;i<n;i++) {
		for (j=i+1;j<n;j++) {
			indexL[i][j]=l;
			indexL[j][i]=l;
			l++;
		}
	}


	// Factorization

	for (i=0;i<n;i++) {   // Columns
		
		outStream << "dl[" << indexD[i][i] << "]=m[" << indexM[i][i] << "]";

		for (k=0;k<i;k++) {
			outStream << "-l[" << indexL[i][k] << "]*dl[" << indexD[i][k] << "]";
		}

		outStream << ";" << endl;
		outStream << "invD[" << i << "]=1/dl[" << indexD[i][i] << "];" << endl;

		
		for (j=0;j<(n-i-1);j++) { 
			outStream << "dl[" << indexD[i][i+j+1] << "]=m["
				<< indexM[i][i+j+1] << "]";
			
			for (k=0;k<i;k++) {
				outStream << "-l[" << indexL[k][i] << "]*dl[" << indexD[k][j+i+1]
					<< "]";
				
			}

			outStream << ";" << endl;

			outStream << "l[" << indexL[i][i+j+1] << "]="
				<< "dl[" << indexD[i][i+j+1] << "]*invD[" 
				<< i << "];" << endl;
		}

	
	}

	// Manipulation of rhs

	outStream << endl;

	for (i=0;i<n;i++) {
		outStream << "g[" << i << "]=b[" << i << "]";
		for (j=0;j<i;j++) {
			outStream << "-l[" << indexL[i][j] << "]*g[" << j << "]";
		}
		outStream << ";" << endl;
	}

	// Calculation of solution

	outStream << endl;

	for (i=n-1;i>-1;i--) {

		outStream << "x[" << i << "]=invD[" << i << "]*(g[" << i << "]"; 
		for (j=n-1;j>i;j--) {
			outStream << "-x[" << j << "]*" << "dl[" << indexD[i][j] << "]";
		}
		outStream << ");" << endl;
	}

	outStream << endl << "}" << endl;

	return(outStream.str());
}



int numberMultiplications(string _string) {
	int i,nm=0;

	for (i=0;i<_string.length();i++) {
		if (_string[i]=='*') {nm++;}
	}

	return(nm);
}

int numberDivisions(string _string) {
	int i,nd=0;

	for (i=0;i<_string.length();i++) {
		if (_string[i]=='/') {nd++;}
	}

	return(nd);
}

int numberAdditions(string _string) {
	int i,na=0;

	for (i=0;i<_string.length();i++) {
		if ((_string[i]=='+') || (_string[i]=='-')) {na++;}
	}

	return(na);
}




